/* common.c
**
**    Copyright (C) 2006  Wei Lian	<lianwei2003.student@sina.com>
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License Version 2 as
**    published by the Free Software Foundation;
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/

#include "header.h"
G_LOCK_DEFINE (title);
G_LOCK_DEFINE (title_len);
G_LOCK_DEFINE (elapsed_seconds);
G_LOCK_DEFINE (flag_buf_rec);

#define CHANNELS_FILE  		"channels.xml"
#define DEFAULT_UPDATE_URL	"http://202.101.35.149/gchlxml"
#define DEFAULT_UPDATE_URL2	"http://www.sopcast.com/gchlxml"

#define MAX_PATH	1024
#define MAX_CHANNELS_FILE_SIZE	10*1024*1024	//10M
#define SOPCAST_HOME	"./.gtk-sopcast"

//////////////network related stuff
int
connect_to_server (char *host, int portnum)
{
  int sock;
  struct sockaddr_in serv_addr;
  struct hostent *server;

  sock = socket (AF_INET, SOCK_STREAM, 0);
  if (sock == -1)
    {
      perror ("ERROR opening socket");
      return -3;
    }
  server = gethostbyname (host);
  if (server == NULL)
    {
      fprintf (stderr, "ERROR, no such host\n");
      return -2;
    }
  memset ((char *) &serv_addr, 0, sizeof (serv_addr));
  serv_addr.sin_family = AF_INET;
  memcpy ((char *) &serv_addr.sin_addr.s_addr,
	  (char *) server->h_addr, server->h_length);
  serv_addr.sin_port = htons (portnum);
  if (connect
      (sock, (struct sockaddr *) &serv_addr, sizeof (serv_addr)) == -1)
    {
      close (sock);
      return -1;
    }
  return sock;

}


int
socket_read (int sock, char *buff, int size)
{
  int red = 0, ret;
  while (size)
    {
      ret = recv (sock, buff + red, size, 0);
      if (ret <= 0)		////ret==0: server shutdown, ret==-1: recv error
	{
	  if (ret == -1)
	    fprintf (stderr, "recv error\n");
	  return ret;
	}
      red += ret;
      size -= ret;
    }
  return 1;
}



///network client using tcp
int
client_read (char *buff, int sizbuff)
{
  int n;
  int sock;
  memset (buff, 0, sizbuff);
  sock = connect_to_server ("127.0.0.1", atoi (outport));
  if (sock < 0)
    return -2;
  ////for sp-sc 0.8, client need to send message first to receive data
  write (sock, "1", strlen ("1") + 1);
  n = recv (sock, buff, sizbuff - 1, 0);
//  n = read (sock, buff, sizbuff - 1);
  if (n == -1)
    {
      perror ("ERROR reading from socket");
      return -1;
    }
  close (sock);
  return 0;
}



/////////////////pthread   

void *
thread_sop_quit ()		///display quit message
{
  gdk_threads_enter ();
  gtk_entry_set_text (GTK_ENTRY (entry_status), "");
  gdk_threads_leave ();

}



void *
thread_connect ()		///display connecting state
{
  char status[64];
  G_LOCK (elapsed_seconds);
  sprintf (status, "Connecting, %d seconds elapsed",
	   (elapsed_seconds++) % 1000);
  G_UNLOCK (elapsed_seconds);

  gdk_threads_enter ();
  gtk_entry_set_text (GTK_ENTRY (entry_status), status);
  gdk_threads_leave ();

}



void *
thread_buf_rec ()
{
  int sock;
  FILE *fp = NULL;
  char *data_path;
  struct timeval t1, t2;
  char t1str[11];
  char word;
  char buffer[1024];
  int i;
  int ret, ret0, red, size;
  fd_set fds;
  struct timeval tv;
  int flag;
  G_LOCK (title_len);
  char *status = malloc (title_len + 64);
  G_UNLOCK (title_len);
  int first_time = 1;

  if (fp != NULL)		///fp serves for double purposes
    {
    CONNECT:
      G_LOCK (title_len);
      data_path = malloc (strlen (record_dir) + title_len + 10 + 4 + 1);
      G_UNLOCK (title_len);
      strcpy (data_path, record_dir);
      strcat (data_path, "/");
      if (access (data_path, F_OK))
	mkdir (data_path, S_IRWXU);
      G_LOCK (title);
      strcat (data_path, title);
      G_UNLOCK (title);
      gettimeofday (&t1, 0);
      sprintf (t1str, "%d", t1.tv_sec);
      strcat (data_path, t1str);
      strcat (data_path, ".asf");
      //////////  
      fp = fopen (data_path, "w");
      free (data_path);
      if (fp == NULL)
	{
	  fprintf (stderr, "\n Cannot open file to record\n");
	  gdk_threads_enter ();
	  gtk_button_set_label (GTK_BUTTON (button_record), "Record");
	  gdk_threads_leave ();
	}
    }
  ///////establish connection
  sock = connect_to_server ("127.0.0.1", atoi (outport));
  if (sock < 0)			///failed to connect to server
    {
      fprintf (stderr, "failed to connect to 127.0.0.1:%s, error=%d\n",
	       outport, sock);
      if (status)
	free (status);
/*      gdk_threads_enter ();
      gtk_entry_set_text (GTK_ENTRY (entry_status), "");
      gdk_threads_leave ();*/

      if (fp != NULL)
	{
	  fclose (fp);
	  gdk_threads_enter ();
	  gtk_button_set_label (GTK_BUTTON (button_record), "Record");
	  gdk_threads_leave ();
	}

      G_LOCK (flag_buf_rec);
      flag_buf_rec = 0;
      G_UNLOCK (flag_buf_rec);
      pthread_exit (NULL);
    }
  ////for sp-sc 0.8, client need to send message first to receive data
  write (sock, "1", strlen ("1") + 1);
  ///////stream header
  gdk_threads_enter ();
  gtk_entry_set_text (GTK_ENTRY (entry_status), "Parsing stream head");
  gdk_threads_leave ();

  for (i = 0; i < 2; i++)
    {
      while (1)
	{
	  ret = recv (sock, &word, 1, 0);
	  if (ret == 1)
	    {
	      if (word == '\n')
		break;
	    }
	  else if (ret == 0 || (ret == -1 && errno != EINTR))	////ret==0: server shutdown, ret==-1: recv error
	    {
	      if (ret == -1)
		perror ("recv");
	      else if (ret == 0)
		fprintf (stderr, "server shutdown while parsing head\n");

	      close (sock);
	      if (status)
		free (status);
/*	      gdk_threads_enter ();
	      gtk_entry_set_text (GTK_ENTRY (entry_status), "");
	      gdk_threads_leave ();*/

	      if (fp != NULL)
		{
		  fclose (fp);
		  gdk_threads_enter ();
		  gtk_button_set_label (GTK_BUTTON (button_record), "Record");
		  gdk_threads_leave ();
		}

	      G_LOCK (flag_buf_rec);
	      flag_buf_rec = 0;
	      G_UNLOCK (flag_buf_rec);

	      pthread_exit (NULL);
	    }
	}
    }

  ////////stream content
  FD_ZERO (&fds);
  while (1)
    {
      gettimeofday (&t1, 0);

      for (i = 0; i < 50; i++)
	{
	  red = 0;
	  size = 1024;
	  while (size)
	    {
	      FD_SET (sock, &fds);
	      tv.tv_sec = 1;
	      tv.tv_usec = 0;
	      ret0 = select (sock + 1, &fds, NULL, NULL, &tv);
	      ////////////
	      if (ret0 == 1)	////got message from server
		{
		  ret = recv (sock, buffer + red, size, 0);
		  if (ret > 0)
		    {
		      red += ret;
		      size -= ret;
		    }
		  else if (ret == 0 || (ret == -1 && errno != EINTR))	////ret==0: server shutdown, ret==-1: recv error
		    {
		      if (ret == -1)
			perror ("recv");
		      else if (ret == 0)
			fprintf (stderr, "server shutdown while streaming\n");

		      close (sock);
		      if (status)
			free (status);
/*		      gdk_threads_enter ();
		      gtk_entry_set_text (GTK_ENTRY (entry_status), "");
		      gdk_threads_leave ();*/

		      if (fp != NULL)
			{
			  fclose (fp);
			  gdk_threads_enter ();
			  gtk_button_set_label (GTK_BUTTON
						(button_record), "Record");
			  gdk_threads_leave ();
			}

		      G_LOCK (flag_buf_rec);
		      flag_buf_rec = 0;
		      G_UNLOCK (flag_buf_rec);

		      pthread_exit (NULL);
		    }
		}
	      else if (ret0 <= 0)	////timeout
		{
		  if (ret == -1)
		    fprintf (stderr, "select failed\n");
//              else
//                fprintf (stderr, "timeout\n");

		  gdk_threads_enter ();
		  flag =
		    strcmp (gtk_button_get_label (GTK_BUTTON (button_record)),
			    "|Quit|");
		  gdk_threads_leave ();
		  if ((flag == 0) && (fp == NULL))	////start recording
		    {
		      close (sock);
//                    fp = 1;
		      goto CONNECT;
		    }
		  if ((flag != 0) && (fp != NULL))	////stop recording
		    {
		      fclose (fp);
		      fp = NULL;
		    }
		}
	      else if (ret0 > 1)
		fprintf (stderr, "select set error\n");
	    }
	  ////////// 
	  gdk_threads_enter ();
	  flag =
	    strcmp (gtk_button_get_label (GTK_BUTTON (button_record)),
		    "|Quit|");
	  gdk_threads_leave ();
	  if ((flag == 0) && (fp == NULL))	////start recording
	    {
	      close (sock);
//            fp = 1;
	      goto CONNECT;
	    }
	  if ((flag != 0) && (fp != NULL))	////stop recording
	    {
	      fclose (fp);
	      fp = NULL;
	    }

	  if (fp != NULL)
	    fwrite (buffer, red, 1, fp);

	  if (first_time)
	    {			////buffering rate
	      if (fp != NULL)
		sprintf (status, "Recording %d%% | ", i + 1);
	      else
		sprintf (status, "Buffering %d%% | ", i + 1);
	      G_LOCK (title);
	      strcat (status, title);
	      G_UNLOCK (title);
	      gdk_threads_enter ();
	      gtk_entry_set_text (GTK_ENTRY (entry_status), status);
	      gdk_threads_leave ();
	    }
	}
//      if (!first_time)
      {				////streamming rate
	gettimeofday (&t2, 0);
	if (fp != NULL)
	  sprintf (status, "Recording %lus | ", (t2.tv_sec - t1.tv_sec));
	else
	  sprintf (status, "Streaming %lus | ", (t2.tv_sec - t1.tv_sec));
	G_LOCK (title);
	strcat (status, title);
	G_UNLOCK (title);
	gdk_threads_enter ();
	gtk_entry_set_text (GTK_ENTRY (entry_status), status);
	gdk_threads_leave ();
      }
      first_time = 0;

    }
}






////////////////////fork new process
void
fork_channel ()
{
  pid_t pid;
  pid = fork ();
  if (pid == -1)
    {
      perror ("fork");
      pid_channel = -2;
      exit (-1);
    }
  else if (pid == 0)
    {
      char data_path[MAX_PATH];
      sprintf(data_path,"%s/%s",getenv ("HOME"),SOPCAST_HOME );
      mkdir (data_path, S_IRWXU);
      sprintf(data_path,"%s/%s/%s",getenv ("HOME"),SOPCAST_HOME,CHANNELS_FILE );
      execlp ("wget", "wget", "-q", channels_url, "-O", data_path, NULL);
      perror ("wget");
      _exit (-1);
    }
  else
    {
      // parent 
      pid_channel = pid;
    }

}


void
fork_sop (char *play_url)
{

  //////////setting inport and outport
  int i;
  char inport[8];
  int sock;

  ///choose inport and outport number
  for (i = 8000; i < 9000; i++)
    {
      sock = connect_to_server ("127.0.0.1", i);
      if (sock == -1)
	break;
      else if (sock >= 0)
	close (sock);
    }
  sprintf (inport, "%d", i);

  for (i = 9000; i < 10000; i++)
    {
      sock = connect_to_server ("127.0.0.1", i);
      if (sock == -1)
	break;
      else if (sock >= 0)
	close (sock);
    }
  sprintf (outport, "%d", i);


  ///////////////////////
  pid_t pid;
  pid = fork ();
  if (pid == -1)
    {
      perror ("fork");
//      pid_sop = -2;
      exit (-1);
    }
  else if (pid == 0)
    {
      execlp ("sp-sc", "sp-sc", play_url, inport, outport, NULL);
      perror ("sp-sc");
      _exit (-1);
    }
  else
    {
      // parent 
      pid_sop = pid;
    }

}



void
fork_play (char *play_url)
{
  pid_t pid;

  pid = fork ();

  if (pid == -1)
    {
      perror ("fork");
      pid_player = -2;
      exit (-1);
    }
  else if (pid == 0)
    {
      /////////////////////get player info
      char *url;
      char *playopt[16];
      int iter = 0;
      char *line;
      char *savept = NULL;

      line = strdup (gtk_entry_get_text (GTK_ENTRY (entry_player)));

      playopt[0] = strtok_r (line, " ,", &savept);
      while ((playopt[++iter] = strtok_r (NULL, " ,", &savept)) != NULL);

      if (play_url == NULL)
	{
	  url = malloc (strlen ("http://localhost:") + strlen (outport) + 1);
	  strcpy (url, "http://localhost:");
	  strcat (url, outport);
	}
      else
	{
	  url = play_url;
	}

      playopt[iter] = url;
      playopt[++iter] = NULL;
      /////////////mask the error output
      int fd = open ("/dev/zero", O_WRONLY);
      close (2);
      dup (fd);

      /////////////execute command
      execvp (playopt[0], playopt);
      perror ("player");
      _exit (-1);
    }
  else
    {
      // parent 
      pid_player = pid;
    }

}




void
sig_alarm_callback ()
{
  char buff_cmp[] =
    { 72, 84, 84, 80, 47, 49, 46, 48, 32, 50, 48, 48, 32, 79, 75, 13,
    10,
    13, 10,
    48, 38, -78, 117, -114, 102, -49, 17, -90, -39, 0,
  };
  char buff[30];		//size could be larger than 30
  int flag;

  client_read (buff, sizeof (buff));
  ///////////////
  if (strcmp (buff, buff_cmp) == 0)	////connected with remote peers, begin streamming
    {
      if (set_ticker (0, 0) == -1)	////stop alarm signal
	{
	  perror ("set_ticker");
	  exit (-1);
	}
      ////////////////start player
      fork_play (NULL);
      ////////////////////start buffering monitor
      G_LOCK (flag_buf_rec);
      flag = flag_buf_rec;
      G_UNLOCK (flag_buf_rec);

      if (flag == 0)		////make sure there is only one thread_buf_rec thread 
	{
	  G_LOCK (flag_buf_rec);
	  flag_buf_rec = 1;
	  G_UNLOCK (flag_buf_rec);

	  pthread_t pt;
	  pthread_attr_t attr;
	  pthread_attr_init (&attr);
	  pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
	  pthread_create (&pt, &attr, thread_buf_rec, NULL);
	}
    }
  else				////not connected
    {
/*      G_LOCK (flag_buf_rec);
      flag = flag_buf_rec;
      G_UNLOCK (flag_buf_rec);

      if (flag == 0)		////make sure this thread does not coexist with thread_buf_rec thread 
	{*/
      pthread_t pt;
      pthread_attr_t attr;
      pthread_attr_init (&attr);
      pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
      pthread_create (&pt, &attr, thread_connect, NULL);
/*	}*/
    }

}



void
start_play ()
{
  if (pid_player == -2 || kill_player == 0)	////pid_player==-2: player has exited, kill_player==0: player killed
    if (pid_sop > 0)
      {

	G_LOCK (elapsed_seconds);
	elapsed_seconds = 0;
	G_UNLOCK (elapsed_seconds);

	if (set_ticker (1, 1000) == -1)	////start alarm signal
	  {
	    perror ("set_ticker");
	    exit (-1);
	  }
      }

}


///////kill process


void
fork_kill_sop ()
{
  if (pid_sop > 0)
    {
//      if (kill ((pid_t) pid_sop, SIGTERM) == -1)
      ////seems like sp-sc can be SIGKILLed without causing any problem 
      kill ((pid_t) pid_sop, SIGKILL);
    }
  // fprintf (stderr, "Sent kill signal to sp-sc %d\n", (int) pid_sop);

}


void
fork_kill_play ()
{
  if (pid_player > 0)
    {
      kill_player = kill ((pid_t) pid_player, SIGTERM);
      if (kill_player == -1)
	kill_player = kill ((pid_t) pid_player, SIGKILL);
    }
  // fprintf (stderr, "Sent kill signal to player %d\n", (int) pid_player);
}

#ifdef  ORG_UPDATE
void
channel_list_update ()
{
  FILE *fp;
  char *data_path;
  /////////////get channel list file name
  data_path =
    malloc (strlen (getenv ("HOME")) + strlen ("/.gtk-sopcast/index.html") + 1);
  strcpy (data_path, getenv ("HOME"));
  strcat (data_path, "/.gtk-sopcast/index.html");
  ////////////////// 
  fp = fopen ((char *) data_path, "r");
  if (fp == NULL)
    {
      fprintf (stderr, "\nCan not open the channel list file! \n");
      exit (-1);
    }
  free (data_path);
///////////////////load channel list
  char *strtok_buf;
  char *line_get = NULL;
  char *str_tok;
  size_t len = 0;

  int row = 0;
  char rowstr[4];
  char *channel_info[9];
  int i;

  for (i = 0; i < 9; i++)
    channel_info[i] = NULL;

  regex_t re1, re2;
  regmatch_t subs[1];
  char *string;

  regcomp (&re1, "<td[^<>\\/]*>", REG_EXTENDED);
  regcomp (&re2, "<\\/td>", REG_EXTENDED);

  gtk_clist_clear (GTK_CLIST (clist_channel));

  while (getline (&line_get, &len, fp) != -1)
    {
      if ((str_tok = strstr (line_get, channels_url_header)) != NULL)
	{
	  strtok_r (str_tok, "\">", &strtok_buf);
	  channel_info[8] = str_tok;
	  /////////////
	  sprintf (rowstr, "%d", (++row) % 1000);
	  channel_info[0] = rowstr;
	  /////////////
	  string = line_get;
	  for (i = 1; i <= 7; i++)
	    {
	      if (regexec (&re1, string, (size_t) 1, subs, 0) == 0)
		channel_info[i] = string + subs[0].rm_eo;
	      else
		goto NEXT;

	      if (regexec (&re2, string, (size_t) 1, subs, 0) == 0)
		{
		  *(string + subs[0].rm_so) = '\0';

		  string = string + subs[0].rm_eo;
		}
	      else
		goto NEXT;
	    }

	  if (channel_info[1] != NULL)
	    {
	      G_LOCK (title_len);
	      title_len =
		(title_len <
		 strlen (channel_info[1])) ? strlen (channel_info[1])
		: title_len;
	      G_UNLOCK (title_len);
	    }

	  gtk_clist_append (GTK_CLIST (clist_channel), channel_info);
	}
    NEXT:
      ;
    }
  gtk_clist_columns_autosize (GTK_CLIST (clist_channel));

  regfree (&re1);
  regfree (&re2);

  if (line_get)
    free (line_get);
  fclose (fp);
  ////////////////////
  G_LOCK (title);
  G_LOCK (title_len);
  char *newtitle = realloc (title, title_len + 1);
  if (newtitle != NULL)
    title = newtitle;
  G_UNLOCK (title_len);
  G_UNLOCK (title);
}
#else
int  read_channel_data(const char * f ,char * dest,size_t dlen )
{
  FILE *fp=NULL;
  int total=0;
  char * ptr = dest;
  int  len =0;

  fp = fopen (f, "r");
  if (!fp )
    {
      fprintf (stderr, "\nCan not open the channel list file! \n");
      return -1;
    }

  len = fread(ptr ,sizeof(char),dlen - total ,fp );
   while(len>0){ 
	ptr+=len;
	total+=len;
  	len = fread(ptr,sizeof(char),dlen - total ,fp );
  }
  if (len <0 )perror(f);
   fclose(fp);
   return total;
}
void
channel_list_update ()
{
  char data_path[MAX_PATH];
  char *data_buf=NULL;
  char * home = getenv("HOME");
  data_buf = malloc(sizeof(char)*MAX_CHANNELS_FILE_SIZE);
  if (!data_buf) return ;

  memset(data_buf,'\0',MAX_CHANNELS_FILE_SIZE);
  if (0>snprintf(data_path,MAX_PATH-1,"%s/%s/%s",!home? ".":home ,SOPCAST_HOME,CHANNELS_FILE) ){
  	goto   clear_data;
  }
  int len=read_channel_data(data_path,data_buf,MAX_CHANNELS_FILE_SIZE-1);
   if (len <0 ){
  	goto   clear_data;
      }    
   data_buf[len]='\0';
  
///////////////////load channel list
  int row = 0;
  char rowstr[4];
  char *channel_info[9];
  int i;

  gtk_clist_clear (GTK_CLIST (clist_channel));
 char * last_ptr = data_buf + len ;
 char * ptr = data_buf;
 char * p = 0;
  char * p2;
  while ( ptr < last_ptr /*&& (*ptr) */)
    {
	  for (i = 0; i < 9; i++)
    		channel_info[i] = NULL;

	//printf("parsing :%128s \n",ptr);

	  sprintf (rowstr, "%d", (++row) % 1000);
	  channel_info[0] = rowstr;

	  p = strstr(ptr,"<channel");
	  if ( !p  ) break;
	   p+=strlen("<channel");
	  if (p >=last_ptr )break;

	   p2 = strstr(p,"<name");
	   if (!p2) break;
	   p = strstr(p2,">");
	   if (!p ) break;

		p2=strstr(p++,"</");
	    if ( !p2)break;
		
		*p2++='\0';
		channel_info[1]=p;
		
	     ptr= strstr(p2,"<stream_type>");
		if (!ptr)break;
		ptr+=strlen("<stream_type>");
		if (ptr >last_ptr)break;

		p2 = strstr(ptr,"</stream_type>");
		if (!p2 )break;
		*p2++='\0';
		channel_info[2]=ptr;
		
		p2 = strstr(p2,"<kbps>");
		if (!p2)break;
		p2+=strlen("<kbps>");
		if (p2 >last_ptr)break;

		ptr = strstr(p2,"</kbps>");
		if(!ptr)break;
		*ptr++='\0';
		channel_info[3]=p2;
		
		//users
		p2=strstr(ptr,"<user_count>");
		if (!p2)break;
		p2+=strlen("<user_count>");		if (p2 >last_ptr)break;

		ptr = strstr(p2,"</user_count>");
		if(!ptr)break;
		*ptr++='\0';if (ptr >last_ptr)break;
		channel_info[4]=p2;

		p2=strstr(ptr,"<qs>");
		if(!p2)break;
		p2+=strlen("<qs>");if (p2 >last_ptr)break;
		ptr=strstr(p2,"</qs>");
		if(!ptr)break;
		*ptr++='\0';if (ptr>last_ptr)break;
		
		channel_info[5]=p2;
	
		p2=strstr(ptr,"<qc>");
		if(!p2)break;
		p2+=strlen("<qc>");if (p2 >last_ptr)break;
		ptr=strstr(p2,"</qc>");
		if(!ptr)break;
		*ptr++='\0';if (ptr >last_ptr)break;
		
		channel_info[6]=p2;

		p2=strstr(ptr,"<start_from>");
		if(!p2)break;
		p2+=strlen("<start_from>");if (p2 >last_ptr)break;
		ptr=strstr(p2,"</start_from>");
		if(!ptr)break;
		*ptr++='\0';if (ptr >last_ptr)break;
		
		channel_info[7]=p2;

		p2=strstr(ptr,"<item>");
		if(!p2)break;
		p2+=strlen("<item>");if (p2 >last_ptr)break;
		ptr=strstr(p2,"</item>");
		if(!ptr)break;
		*ptr++='\0';
	    channel_info[8] = p2;


	    if(!strcmp(p2,"null")||!strlen(p2) )continue;
	  if (channel_info[1] != NULL)
	    {
	      G_LOCK (title_len);
	      title_len =(title_len <strlen (channel_info[1])) ? strlen (channel_info[1]): title_len;
	      G_UNLOCK (title_len);
	    }
	 for( i=0;i<9;i++){
		printf("channel[%d][%d]=%s \n", row-1,i,channel_info[i]);
	}
	  gtk_clist_append (GTK_CLIST (clist_channel), channel_info);
	
    }//end while
  gtk_clist_columns_autosize (GTK_CLIST (clist_channel));

  ////////////////////
  G_LOCK (title);
  G_LOCK (title_len);
  char *newtitle = realloc (title, title_len + 1);
  if (newtitle != NULL)
    title = newtitle;
  G_UNLOCK (title_len);
  G_UNLOCK (title);

 clear_data:
   if ( data_buf) free(data_buf);

}


#endif



void
sig_child_callback (int sign)
{
  pid_t result;
  int flag;
  while (1)
    {
      result = waitpid (-1, NULL, WNOHANG);
      if (result == 0)
	break;
      if (result == -1)
	{
	  if (errno != EINTR)
	    break;
	  else
	    strerror (errno);
	}
      ////////////////////
      if (result == pid_sop)	//if the value of pid_sop has been updated first, this condition won't be met.
	{
	  pid_sop = -2;		/////reset to default value

	  if (set_ticker (0, 0) == -1)	////stop alarm signal
	    {
	      perror ("set_ticker");
	      exit (-1);
	    }
	  /////////////display quit message
/*	  G_LOCK (flag_buf_rec);
	  flag = flag_buf_rec;
	  G_UNLOCK (flag_buf_rec);
	  if (flag == 0)	////make sure this thread does not coexist with thread_buf_rec thread 
	    {*/
	  pthread_t pt;
	  pthread_attr_t attr;
	  pthread_attr_init (&attr);
	  pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
	  pthread_create (&pt, &attr, thread_sop_quit, NULL);
/*	    }*/
	}
      else if (result == pid_channel)
	{
	  pid_channel = -2;	/////reset to default value
	  channel_list_update ();
	  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON
					(button_channel), FALSE);
	}
      else if (result == pid_player)
	pid_player = -2;
    }

}




void
get_channel_list (int flag)
{
  char data_path[MAX_PATH];
  /////////////get channel list file name
  sprintf(data_path, "%s/%s/%s",getenv ("HOME"),SOPCAST_HOME,CHANNELS_FILE );

  if ((flag || access (data_path, F_OK)) && pid_channel == -2)
    fork_channel ();
  if (flag == 0 && !(access (data_path, F_OK)) && pid_channel == -2)
    channel_list_update ();
}



/////////////////////hareware related stuff
int
sound_read (void)
{
  int result;
  if (ioctl (sound_handle, MIXER_READ (4), &result) == -1)	//read sound volume
    {
      perror ("read_error");
      return;
    }
  result = ((result >> 8) + (result & 0xFF)) >> 1;
  result = (result > 0) ? result : 0;
  result = (result < 100) ? result : 100;
  return result;
}


void
sound_write (int leftright)
{
  leftright = (leftright < 0) ? 0 : leftright;
  leftright = (leftright > 100) ? 100 : leftright;
  leftright = (leftright << 8) | (leftright & 0xFF);
  if (ioctl (sound_handle, MIXER_WRITE (4), &leftright) == -1)	//write sound wolume
    {
      perror ("write_error");
      return;
    }
}



int
set_ticker (int value, int interval)
{

  struct itimerval new_tim;
  long n_se, n_us;

  n_se = interval / 1000;
  n_us = (interval % 1000) * 1000L;
  new_tim.it_interval.tv_sec = n_se;
  new_tim.it_interval.tv_usec = n_us;

  n_se = value / 1000;
  n_us = (value % 1000) * 1000L;
  new_tim.it_value.tv_sec = n_se;
  new_tim.it_value.tv_usec = n_us;

  return setitimer (ITIMER_REAL, &new_tim, NULL);
}
